(* $Id: gregorianDate.ml,v 1.1 2004/10/24 18:45:06 rich Exp $
 * Gregorian date calculations, inspired by and derived from Perl's
 * Date::Calc.
 * Copyright (C) 2004 Merjis Ltd. [OCaml version]
 * Copyright (C) 1995 - 2002 by Steffen Beyer [Date::Calc]
 *)

let leap_year year =
  year land 0x3 = 0 &&
  (let y100 = year / 100 in
   y100 * 100 <> year || y100 land 0x3 = 0)

let index_of_bool = function
  | false -> 0
  | true -> 1

let days_in_year_ =
  [|
    [| 0; 0; 31; 59; 90; 120; 151; 181; 212; 243; 273; 304; 334; 365 |];
    [| 0; 0; 31; 60; 91; 121; 152; 182; 213; 244; 274; 305; 335; 366 |];
  |]

let days_in_year ?(month = 12) year =
  if year > 0 then (
    if month >= 1 && month <= 12 then
      days_in_year_.(index_of_bool (leap_year year)).(month+1)
    else
      invalid_arg "days_in_year: bad month"
  ) else
    invalid_arg "days_in_year: bad year"

let days_in_month_ =
  [|
    [| 0; 31; 28; 31; 30; 31; 30; 31; 31; 30; 31; 30; 31 |];
    [| 0; 31; 29; 31; 30; 31; 30; 31; 31; 30; 31; 30; 31 |];
  |]

let days_in_month year month =
  if year > 0 then (
    if month >= 1 && month <= 12 then
      days_in_month_.(index_of_bool (leap_year year)).(month)
    else
      invalid_arg "days_in_month: bad month"
  ) else
    invalid_arg "days_in_month: bad year"

let year_to_days year =
  let days = year * 365 in
  let year = year lsr 2 in
  let days = days + year in
  let year = year / 25 in
  let days = days - year in
  let days = days + (year lsr 2) in
  days

let date_to_days (year, month, day) =
  if year >= 1 && month >= 1 && month <= 12 && day >= 1 then (
    let leap = index_of_bool (leap_year year) in
    if day <= days_in_month_.(leap).(month) then
      year_to_days (year-1) + days_in_year_.(leap).(month) + day
    else
      invalid_arg "date_to_days: bad day"
  ) else
    invalid_arg "date_to_days: bad month or year"

let delta_days d1 d2 =
  date_to_days d2 - date_to_days d1

let add_delta_days date dd =
  let days = date_to_days date in
  let days = days + dd in
  if days <= 0 then
    invalid_arg "add_delta_days: invalid resulting date";

  let year = int_of_float (float days /. 365.2425) in
  let day = days - year_to_days year in
  let year, day =
    if day < 1 then (
      let year = year-1 in
      year, days - year_to_days year
    ) else
      year+1, day in
  let leap = index_of_bool (leap_year year) in
  let year, leap, day =
    if day > days_in_year_.(leap).(13) then (
      let day = day - days_in_year_.(leap).(13) in
      let year = year+1 in
      let leap = index_of_bool (leap_year year) in
      year, leap, day
    ) else
      year, leap, day in
  let rec loop month day =
    if day > days_in_year_.(leap).(month) then (
      let day = day - days_in_year_.(leap).(month) in
      month, day
    ) else if month = 1 then
      month, day
    else
      loop (month-1) day
  in
  let month, day = loop 12 day in
  year, month, day

let day_of_week date =
  let days = date_to_days date in
  let days = days - 1 in
  let days = days mod 7 in
  let days = days + 1 in
  days

let weeks_in_year year =
  if year > 0 then
    52 + (index_of_bool (day_of_week (year, 1, 1) = 4 ||
			 day_of_week (year, 12, 31) = 4))
  else
    invalid_arg "weeks_in_year: bad year"

let check_date (year, month, day) =
  year >= 1 &&
  month >= 1 && month <= 12 &&
  day >= 1 && day <= days_in_month_.(index_of_bool (leap_year year)).(month)

let check_time (hour, min, sec) =
  hour >= 0 && min >= 0 && sec >= 0 &&
  hour < 24 && min < 60 && sec < 60

let check_business_date (year, week) dow =
  year >= 1 &&
  week >= 1 && week <= weeks_in_year year &&
  dow >= 1 && dow <= 7

let day_of_year (year, month, day) =
  if year >= 1 && month >= 1 && month <= 12 && day >= 1 then (
    let leap = index_of_bool (leap_year year) in
    if day <= days_in_month_.(leap).(month) then
      days_in_year_.(leap).(month) + day
    else
      invalid_arg "day_of_year: bad day"
  ) else
    invalid_arg "day_of_year: bad month or year"

let week_number (year, month, day) =
  let first = day_of_week (year, 1, 1) - 1 in
  (delta_days (year, 1, 1) (year, month, day) + first) / 7
    + index_of_bool (first < 4)

let week_of_year (year, month, day) =
  if not (check_date (year, month, day)) then
    invalid_arg "week_of_year: bad date";
  let week = week_number (year, month, day) in
  if week = 0 then (
    let year = year - 1 in
    year, weeks_in_year year
  ) else if week > weeks_in_year year then
    year+1, 1
  else
    year, week

let monday_of_week (year, week) =
  let first = day_of_week (year, 1, 1) - 1 in
  let week = if first < 4 then week-1 else week in
  add_delta_days (year, 1, 1) (week * 7 - first)

let nth_weekday_of_month (year, month) dow n =
  if year < 1 || month < 1 || month > 12 || dow < 1 || dow > 7 ||
    n < 1 || n > 5 then
      invalid_arg "nth_weekday_of_month: bad date";
  let first = day_of_week (year, month, 1) in
  let dow = if dow < first then dow + 7 else dow in
  let delta = dow - first in
  let delta = delta + (n-1) * 7 in
  let year, month', day = add_delta_days (year, month, 1) delta in
  if month = month' then
    (year, month, day)
  else
    (* This weekday doesn't exist in this month, eg. 5th Friday in Feb '04 *)
    raise Not_found

let business_of_standard ((year, month, day) as date) =
  let year, week = week_of_year date in
  let dow = day_of_week date in
  (year, week), dow

let standard_of_business ((year, week) as date) dow =
  if not (check_business_date date dow) then
    invalid_arg "standard_of_business: invalid business date";
  let first = day_of_week (year, 1, 1) in
  let delta = ((week + (index_of_bool (first > 4)) - 1) * 7) + (dow - first) in
  add_delta_days (year, 1, 1) delta

let easter_sunday year =
  if year < 1583 || year > 2299 then
    invalid_arg "easter_sunday: cannot compute Easter date for this year";

  let m, n =
    if year < 1700 then 22, 2
    else if year < 1800 then 23, 3
    else if year < 1900 then 23, 4
    else if year < 2100 then 24, 5
    else if year < 2200 then 24, 6
    else 25, 0 in

  let a = year mod 19 in
  let b = year mod 4 in
  let c = year mod 7 in
  let d = (19 * a + m) mod 30 in
  let e = (2 * b + 4 * c + 6 * d + n) mod 7 in
  let day = 22 + d + e in
  let month = 3 in
  let day, month =
    if day > 31 then
      day-31, month+1
    else
      day, month in
  let day =
    if day = 26 && month = 4 then 19 else day in
  let day =
    if day = 25 && month = 4 && d = 28 && e = 6 && a > 10 then 18 else day in
  (year, month, day)
